module code.devicecalculus.Semantics where

open import Data.List hiding (unfold ; [_])
open import Data.Nat hiding (_<_)
open import Data.Bool
open import Data.Product
open import Data.Sum

open import Data.Fin renaming (zero to ZERO; suc to SUC) hiding (lift)
open import Data.Vec renaming (lookup to vecref)

open import code.common.Contexts
open import code.common.Idioms 
open import code.common.MachineWords hiding (_⊠_)

open import code.devicecalculus.BitPatterns renaming (match to bitmatch)
open import code.devicecalculus.Types
open import code.devicecalculus.Syntax

open import code.devicecalculus.MealyMachine

---
--- Environments
---

Env : Cxt Type → Set
Env Γ = All Value Γ

---
--- Dynamic Semantics
---

infix 3  ⟦_⟧

--
-- Evaluator for _⊩ᵈ_
--
⟦_⟧ : ∀ {Γ t} → (Γ ⊩ᵈ t) → Env Γ → Value t
⟦ var x ⟧ ρ       = lookup∈ ρ x
⟦ lam f ⟧ ρ       = λ s → ⟦ f ⟧ (s ▷ ρ) 
⟦ app f e ⟧ ρ     = (⟦ f ⟧ ρ) (⟦ e ⟧ ρ)
--
-- Products and Sums
--
⟦ a ⊠ b ⟧ ρ    = ⟦ a ⟧ ρ , ⟦ b ⟧ ρ
⟦ π1 e ⟧ ρ        = proj₁ (⟦ e ⟧ ρ)
⟦ π2 e ⟧ ρ        = proj₂ (⟦ e ⟧ ρ)
⟦ inl e ⟧ ρ       = inj₁ (⟦ e ⟧ ρ)
⟦ inr e ⟧ ρ       = inj₂ (⟦ e ⟧ ρ)
⟦ 𝟙 ⟧ ρ           = tt
-- Bit vector operations
⟦ 0! ⟧ ρ          = false
⟦ 1! ⟧ ρ          = true
⟦ ⟨ v ⟩ ⟧ ρ          = v
⟦ ⦉ n ⦊ ⟧ ρ          = n
⟦ tlΛ m {n} ⟧ ρ      = λ x → Data.Vec.drop n x
⟦ e₁ && e₂ ⟧ ρ       = bitwise-and (⟦ e₁ ⟧ ρ) (⟦ e₂ ⟧ ρ)
⟦ caseᵇᵛ e p f b ⟧ ρ = (⟦ f ⟧ ρ) $ (bitmatch p (⟦ e ⟧ ρ)) , (⟦ b ⟧ ρ)
⟦ e [ n ] ⟧ ρ       = (⟦ e ⟧ ρ) ⇂ (⟦ n ⟧ ρ)
  where
    -- safe vector projection
    _⇂_ : ∀ {n} {A : Set} → Vec A n → (i : Fin n) → A
    [] ⇂ ()
    (x ∷ xs) ⇂ ZERO  = x
    (x ∷ xs) ⇂ SUC i = xs ⇂ i
⟦ e₁ ⧾ e₂ ⟧ ρ       = ⟦ e₁ ⟧ ρ ⊞ ⟦ e₂ ⟧ ρ
⟦ e₁ ⧿ e₂ ⟧ ρ      = ⟦ e₁ ⟧ ρ ⊟ ⟦ e₂ ⟧ ρ
⟦ e₁ * e₂ ⟧ ρ      = (⟦ e₁ ⟧ ρ) code.common.MachineWords.⊠ (⟦ e₂ ⟧ ρ)
⟦ lit n ⟧ ρ       = n
⟦ true ⟧ ρ        = true
⟦ false ⟧ ρ       = false
⟦ less x y ⟧ ρ    = ⟦ x ⟧ ρ lt ⟦ y ⟧ ρ
  where
    _lt_ : ℕ → ℕ → Bool
    0 lt 0         = false
    0 lt suc n     = true
    suc n lt 0     = false
    suc n lt suc m = n lt m
⟦ plus e1 e2 ⟧ ρ  = ⟦ e1 ⟧ ρ Data.Nat.+ ⟦ e2 ⟧ ρ
⟦ if p x y ⟧ ρ    = if ⟦ p ⟧ ρ then ⟦ x ⟧ ρ else ⟦ y ⟧ ρ
⟦ unfoldDev f i s o ⟧ ρ = unfoldM (⟦ f ⟧ ρ) (⟦ i ⟧ ρ) (⟦ s ⟧ ρ) (⟦ o ⟧ ρ) 
⟦ d1 || d2 ⟧ ρ          = ⟦ d1 ⟧ ρ ∥ ⟦ d2 ⟧ ρ
⟦ feedback θ c o d ⟧ ρ  = feedbackM (⟦ θ ⟧ ρ) (⟦ c ⟧ ρ) (⟦ o ⟧ ρ) (⟦ d ⟧ ρ) 
