module code.devicecalculus.Syntax where

open import Data.Fin renaming (zero to ZERO; suc to SUC) hiding (lift)
open import Data.List hiding (unfold ; [_])
open import Data.Nat hiding (_<_) renaming (_*_ to _<*>_)
open import Data.Vec renaming (lookup to vecref)

open import code.common.Contexts

open import code.devicecalculus.BitPatterns renaming (match to matchbv)
open import code.common.MachineWords renaming (zero to machine0) hiding (_⊠_)
open import code.devicecalculus.Types

infixr 4 _⊠_
infix 3 _⊩ᵈ_

data _⊩ᵈ_ : Cxt Type → Type → Set where
  var     : ∀ {Γ a} → a ∈ Γ → Γ ⊩ᵈ a
  lam     : ∀ {Γ a b} → (a ∷ Γ) ⊩ᵈ b → Γ ⊩ᵈ (a ⇒ b)
  app     : ∀ {Γ a b} → Γ ⊩ᵈ (a ⇒ b) → Γ ⊩ᵈ a → Γ ⊩ᵈ b
  lit     : ∀ {Γ} → (n : ℕ) → Γ ⊩ᵈ nat
  true    : ∀ {Γ} → Γ ⊩ᵈ bool
  false   : ∀ {Γ} → Γ ⊩ᵈ bool
  less    : ∀ {Γ} → (a b : Γ ⊩ᵈ nat) → Γ ⊩ᵈ bool
  plus    : ∀ {Γ} → (a b : Γ ⊩ᵈ nat) → Γ ⊩ᵈ nat
  if      : ∀ {Γ} → {t : Type} (a : Γ ⊩ᵈ bool) (b c : Γ ⊩ᵈ t) → Γ ⊩ᵈ t
  -- products and sums
  _⊠_    : ∀ {Γ} {t1 t2 : Type} (a : Γ ⊩ᵈ t1) (b : Γ ⊩ᵈ t2) → Γ ⊩ᵈ (t1 ⊗ t2)
  π1      : ∀ {Γ} {t1 t2 : Type} (a : Γ ⊩ᵈ (t1 ⊗ t2)) → (Γ ⊩ᵈ t1)
  π2      : ∀ {Γ} {t1 t2 : Type} (b : Γ ⊩ᵈ (t1 ⊗ t2)) → (Γ ⊩ᵈ t2)
  inl     : ∀ {Γ} {t1 t2 : Type} (a : Γ ⊩ᵈ t1) → Γ ⊩ᵈ (t1 ⊕ t2)
  inr     : ∀ {Γ} {t1 t2 : Type} (b : Γ ⊩ᵈ t2) → Γ ⊩ᵈ (t1 ⊕ t2)
  𝟙       : ∀ {Γ} → Γ ⊩ᵈ unit
  -- Bit level operations
  0!      : ∀ {Γ} → Γ ⊩ᵈ bit
  1!      : ∀ {Γ} → Γ ⊩ᵈ bit
  -- standard logic vectors
  ⟨_⟩    : ∀ {Γ n} → Vec Bit n → Γ ⊩ᵈ (slv n)
  ⦉_⦊    : ∀ {Γ n} → Fin n → Γ ⊩ᵈ (index n)
  caseᵇᵛ : ∀ {Γ n c} → Γ ⊩ᵈ (slv n) → Vec BitPat n → Γ ⊩ᵈ (slv n ⇒ c) → Γ ⊩ᵈ c → Γ ⊩ᵈ c
  _[_]  : ∀ {Γ n} → Γ ⊩ᵈ (slv n) → Γ ⊩ᵈ (index n) → Γ ⊩ᵈ bit
  _⧾_     : ∀ {Γ n} → Γ ⊩ᵈ (slv n) → Γ ⊩ᵈ (slv n) → Γ ⊩ᵈ (slv n)
  _⧿_     : ∀ {Γ n} → Γ ⊩ᵈ (slv n) → Γ ⊩ᵈ (slv n) → Γ ⊩ᵈ (slv n)
  _*_     : ∀ {Γ n} → Γ ⊩ᵈ (slv n) → Γ ⊩ᵈ (slv n) → Γ ⊩ᵈ (slv n)

  tlΛ   : ∀ {Γ} m {n} → Γ ⊩ᵈ (slv (n Data.Nat.+ m) ⇒ slv m)
  _&&_  : ∀ {Γ n} → Γ ⊩ᵈ (slv n) → Γ ⊩ᵈ (slv n) → Γ ⊩ᵈ (slv n)
  
  --
  -- Corresponds directly to the Mealy constructor:
  --    unfoldM : ∀ {i s o} → (i → s → (o × s)) → i → s → o → Mealy i s o
  --
  unfoldDev : ∀ {Γ i s o} →
                Γ ⊩ᵈ (i ⇒ s ⇒ (o ⊗ s)) →
                Γ ⊩ᵈ i                  →
                Γ ⊩ᵈ s                  →
                Γ ⊩ᵈ o                  →
                -----------------------
                Γ ⊩ᵈ Dev i s o
  _||_     : ∀ {i1 s1 o1 i2 s2 o2 Γ}
           → Γ ⊩ᵈ Dev i1 s1 o1       
           → Γ ⊩ᵈ Dev i2 s2 o2     
             ---------------------------------------             
           → Γ ⊩ᵈ Dev (i1 ⊗ i2) (s1 ⊗ s2) (o1 ⊗ o2)
  feedback : ∀ {i1 i2 s o1 o2 Γ}  
           → Γ ⊩ᵈ (o1 ⇒ o2)      
           → Γ ⊩ᵈ (i2 ⇒ o1 ⇒ i1) 
           → Γ ⊩ᵈ i2             
           → Γ ⊩ᵈ Dev i1 s o1    
             ----------------
           → Γ ⊩ᵈ Dev i2 s o2        


--
-- Follows Wadler's https://plfa.github.io/DeBruijn/
--
ext : ∀ {Γ Γ'} →
        (∀ {a : Type} → a ∈ Γ → a ∈ Γ') →
        --------------------------------------
        (∀ {a b} → a ∈ (b ∷ Γ) → a ∈ (b ∷ Γ'))
ext ρ zero    = zero
ext ρ (suc x) = suc (ρ x)

rename : ∀ {Γ Γ'} →
           (∀ {a} → a ∈ Γ → a ∈ Γ') →
           (∀ {a} → Γ ⊩ᵈ a → Γ' ⊩ᵈ a)
rename ρ (var x)           = var (ρ x)
rename ρ (lam e)           = lam (rename (ext ρ) e)
rename ρ (app e1 e2)       = app (rename ρ e1) (rename ρ e2)
rename ρ (lit n)           = lit n
rename ρ true              = true
rename ρ false             = false
rename ρ (less e1 e2)      = less (rename ρ e1) (rename ρ e2)
rename ρ (plus e1 e2)      = plus (rename ρ e1) (rename ρ e2)
rename ρ (if e1 e2 e3)     = if (rename ρ e1) (rename ρ e2) (rename ρ e3)
rename ρ (e1 ⊠ e2)         = rename ρ e1 ⊠ rename ρ e2
rename ρ (π1 e)            = π1 (rename ρ e)
rename ρ (π2 e)            = π2 (rename ρ e)
rename ρ (inl e)           = inl (rename ρ e)
rename ρ (inr e)           = inr (rename ρ e) 
rename ρ 𝟙                 = 𝟙
rename ρ 0!                = 0!
rename ρ 1!                = 1!
rename ρ ⟨ x ⟩             = ⟨ x ⟩
rename ρ ⦉ x ⦊             = ⦉ x ⦊
rename ρ (caseᵇᵛ e x e1 e2) = caseᵇᵛ (rename ρ e) x (rename ρ e1) (rename ρ e2)
rename ρ (e1 [ e2 ])       = _[_] (rename ρ e1) (rename ρ e2)
rename ρ (e1 ⧾ e2)         = rename ρ e1 ⧾ rename ρ e2
rename ρ (e1 ⧿ e2)         = rename ρ e1 ⧿ rename ρ e2
rename ρ (e1 * e2)         = rename ρ e1 * rename ρ e2 
rename ρ (tlΛ m)             = tlΛ m
rename ρ (e1 && e2)          = rename ρ e1 && rename ρ e2 
rename ρ (unfoldDev f i s o) = unfoldDev (rename ρ f) (rename ρ i) (rename ρ s) (rename ρ o)
rename ρ (d1 || d2)          = rename ρ d1 || rename ρ d2 
rename ρ (feedback θ c o d)  = feedback (rename ρ θ) (rename ρ c) (rename ρ o) (rename ρ d)

push1 : ∀ {Γ x} → (∀ {a : Type} → a ∈ Γ → a ∈ (x ∷ Γ))
push1 x = suc x

push2 : ∀ {Γ x y} → (∀ {a : Type} → a ∈ Γ → a ∈ (x ∷ y ∷ Γ))
push2 x = suc (suc x)

push3 : ∀ {Γ x y z} → (∀ {a : Type} → a ∈ Γ → a ∈ (x ∷ y ∷ z ∷ Γ))
push3 x = suc (suc (suc x))
