module code.examples.Combinational where

-- revisit the imports
open import Agda.Builtin.String renaming (primStringAppend to _+++_)
open import Agda.Primitive
open import Data.Bool renaming (Bool to Bit; false to 0#; true to 1#)
open import Data.Fin using (Fin; zero; suc)
open import Data.List hiding (lookup ; [_] ; and)
open import Data.Maybe hiding (map)
open import Data.Nat using (ℕ; _∸_; _<_ ; _≤_ ; _+_) renaming (zero to Z; suc to S)
open import Data.Product hiding (map ; swap ; curry ; uncurry)
open import Data.Vec hiding (foldl ; foldr ; _++_ ; map ; _[_]=_ ; length) renaming (lookup to vecref ; reverse to rev)

open import code.common.Idioms hiding (_∘_)
open import code.common.Contexts
open import code.common.MachineWords
open import code.common.Stream renaming (hd to shd; tl to stl ; map to mapStr ; zip to zipStr)

open import code.firrtl.Types
open import code.firrtl.Syntax
open import code.firrtl.TransCircuit

open import code.devicecalculus.Types
open import code.devicecalculus.Syntax
open import code.devicecalculus.MealyMachine
open import code.devicecalculus.Semantics


-- Some abbreviations

W1 : Ty
W1 = ↑ UInt 1

W16 : Ty
W16 = ↑ UInt 16

-- circuit Combinational : @[:@2.0]
--   module Combinational : @[:@3.2]
--     input clock : Clock @[:@4.4]
--     input reset : UInt<1> @[:@5.4]
--     input io_x : UInt<16> @[:@6.4]
--     input io_y : UInt<16> @[:@6.4]
--     output io_z : UInt<16> @[:@6.4]

--     node _T_11 = add(io_x, io_y) @[Combinational.scala 12:16:@8.4]
--     node _T_12 = tail(_T_11, 1) @[Combinational.scala 12:16:@9.4]
--     io_z <= _T_12 @[Combinational.scala 12:8:@10.4]
--     skip

Δ : ModTy
Δ = Mod (W1 ⊗ W16 ⊗ W16 ⊗ Unit) , Unit , (W16 ⊗ Unit)  ⋊ []

io_z : [] ⊢ₓ UInt 16 // Δ 
io_z = add io_x io_y
  where
    io_x : [] ⊢ₓ UInt 16 // Δ
    io_x = inp (Sᵍ Zᵍ)
    io_y : [] ⊢ₓ UInt 16 // Δ
    io_y = inp (Sᵍ (Sᵍ Zᵍ))

comb_mod : [] ⊢ᵐ Δ
comb_mod = mod ◇
                (node none (⊗case io_z unit))
                (node none unit)
                (node none unit)

-- The encoding
combinational : [] ⊢ᶜ Δ
combinational = mod comb_mod

combinationalTr : [] ⊩ᵈ Dev (is Δ) (rs Δ) (os Δ)
combinationalTr = trC Δ [] combinational

combinational' : Mealy (Vec Bit 1 × Vec Bit 16 × Vec Bit 16 × ⊤) ⊤ (Vec Bit 16 × ⊤)
combinational' = ⟦ combinationalTr ⟧ []

-- |
-- | Testing
-- |

zero1 : Vec Bit 1
zero1 = 0# ∷ []

-- |
-- | The application of reverse (rev) to these bitstrings is because the arithmetic operations
-- | defined in MachineWords.agda operate on words arranged from least- to most-significant bit.
-- | 

zero16 : Vec Bit 16
zero16 = rev (0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ [])

one16 : Vec Bit 16
one16 = rev (0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 1# ∷ [])

two16 : Vec Bit 16
two16 = rev (0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 1# ∷ 0# ∷ [])

three16 : Vec Bit 16
three16 = rev  (0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 1# ∷ 1# ∷ [])

four16 : Vec Bit 16
four16 = rev  (0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 1# ∷ 0# ∷ 0# ∷ [])

five16 : Vec Bit 16
five16 = rev  (0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 1# ∷ 0# ∷ 1# ∷ [])

six16 : Vec Bit 16
six16 = rev  (0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 1# ∷ 1# ∷ 0# ∷ [])

seven16 : Vec Bit 16
seven16 = rev  (0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 1# ∷ 1# ∷ 1# ∷ [])

eight16 : Vec Bit 16
eight16 = rev  (0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 1# ∷ 0# ∷ 0# ∷ 0# ∷ [])

nine16 : Vec Bit 16
nine16 = rev  (0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 1# ∷ 0# ∷ 0# ∷ 1# ∷ [])

ten16 : Vec Bit 16
ten16 = rev  (0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 1# ∷ 0# ∷ 1# ∷ 0# ∷ [])

-- | 
-- | These match the tests from Combinational.{lo.fir,inputs,outputs}
-- |
testinputs : Stream (Vec Bit 1 × Vec Bit 16 × Vec Bit 16 × ⊤ {lzero})
testinputs = cycle testhd testtl
   where
     testhd : (Vec Bit 1 × Vec Bit 16 × Vec Bit 16 × ⊤ {lzero})
     testhd = (zero1 , zero16 , zero16 , tt)  
     testtl : List (Vec Bit 1 × Vec Bit 16 × Vec Bit 16 × ⊤ {lzero})
     testtl = (zero1 , one16  , two16 , tt)    ∷
              (zero1 , two16  , two16 , tt)    ∷
              (zero1 , two16  , three16 , tt)  ∷
              (zero1 , four16 , four16 , tt)   ∷
              (zero1 , four16 , five16 , tt)   ∷
              (zero1 , six16 , seven16 , tt)   ∷
              (zero1 , seven16 , seven16 , tt) ∷
              (zero1 , eight16 , nine16 , tt)  ∷
              (zero1 , ten16 , ten16 , tt)     ∷ []

triple : String → String → String → String
triple i s o = "(" +++ (i +++ (" , " +++ (s +++ (" , " +++ (o +++ ")")))))

-- hideous
pp : ((Vec Bit 1 × Vec Bit 16 × Vec Bit 16 × ⊤ {lzero}) × ⊤ {lzero} × Vec Bit 16 × ⊤ {lzero}) → String
pp ( ( r , io_x , io_y , _) , _ , (io_z , _)) = triple ppi "tt" (pp16 io_z)
  where
    ppi = triple (ppBV r) (pp16 io_x) (pp16 io_y)
    
-- | Matches the simulator exactly.
runcombinational = Data.List.tail (map pp (takeStr 11 (transcript combinational' testinputs)))

{-
"((0 , 0000000000000000 , 0000000000000000) , tt , 0000000000000000)" ∷
"((0 , 0000000000000001 , 0000000000000010) , tt , 0000000000000011)" ∷
"((0 , 0000000000000010 , 0000000000000010) , tt , 0000000000000100)" ∷
"((0 , 0000000000000010 , 0000000000000011) , tt , 0000000000000101)" ∷
"((0 , 0000000000000100 , 0000000000000100) , tt , 0000000000001000)" ∷
"((0 , 0000000000000100 , 0000000000000101) , tt , 0000000000001001)" ∷
"((0 , 0000000000000110 , 0000000000000111) , tt , 0000000000001101)" ∷
"((0 , 0000000000000111 , 0000000000000111) , tt , 0000000000001110)" ∷
"((0 , 0000000000001000 , 0000000000001001) , tt , 0000000000010001)" ∷ 
"((0 , 0000000000001010 , 0000000000001010) , tt , 0000000000010100)" ∷ []
-}
