module code.examples.Figure2 where

open import Agda.Builtin.String renaming (primStringAppend to _+++_)
open import Data.List
open import Data.Product

open import code.common.MachineWords
open import code.common.Stream renaming (hd to shd; tl to stl ; map to mapStr ; zip to zipStr)
open import code.devicecalculus.MealyMachine

--
-- Example from Figure 2.
--


W4 = BitVector 4

genfib4 : (Bit × Bit) → W4 × W4 → (W4 × (W4 × W4))
genfib4 (_ , 1#) (_ , _) = (v0 , (v0 , v1))
genfib4 (0# , _) (n , m) = (n , (n , m))
genfib4 (1# , _) (n , m) = (m , (m , n ⊞ m))

fib4 : Mealy (Bit × Bit) (W4 × W4) W4
fib4 = unfoldM genfib4 (0# , 0#) (v0 , v0) v0

seed1 : List (Bit × Bit)
seed1 =             (1# , 0#) ∷
        (0# , 0#) ∷ (1# , 0#) ∷ 
        (0# , 0#) ∷ (1# , 0#) ∷ 
        (0# , 0#) ∷ (1# , 0#) ∷ 
        (0# , 1#) ∷ (1# , 0#) ∷ 
        (0# , 0#) ∷ (1# , 0#) ∷ 
        (0# , 0#) ∷ (1# , 0#) ∷ 
        (0# , 0#) ∷ (1# , 0#) ∷ 
        (0# , 0#) ∷ (1# , 0#) ∷ 
        (0# , 0#) ∷ (1# , 1#) ∷ []


ppBit : Bit → String
ppBit b = primShowChar (bit2char b)

pp4 : ((Bit × Bit) × (W4 × W4) × W4) → String
pp4 ((b1 , b2) , (w1 , w2) , w3 ) = "(" +++ (ppBit b1 +++ (" , " +++ (ppBit b2 +++ (") , (" +++
                                    (ppBV w1 +++ (" , " +++ (ppBV w2 +++ (") , " +++
                                    (ppBV w3 +++ ")")))))))))

-------------------------------------------------------------
-------------------------------------------------------------
-- Below examples from the paper for Figure 2; seed0 defined
-- in common/Stream
-------------------------------------------------------------
-------------------------------------------------------------

figure2 : List String
figure2 = Data.List.map pp4 (runM 12 fib4 (cycle (0# , 1#) seed0))

{-
"('0' , '0') , (0000 , 0000) , 0000)" ∷
"('0' , '1') , (0000 , 0001) , 0000)" ∷
"('1' , '0') , (0001 , 0001) , 0001)" ∷
"('0' , '0') , (0001 , 0001) , 0001)" ∷
"('1' , '0') , (0001 , 0010) , 0001)" ∷
"('0' , '0') , (0001 , 0010) , 0001)" ∷
"('1' , '0') , (0010 , 0011) , 0010)" ∷
"('0' , '0') , (0010 , 0011) , 0010)" ∷
"('1' , '0') , (0011 , 0101) , 0011)" ∷
"('0' , '0') , (0011 , 0101) , 0011)" ∷
"('1' , '0') , (0101 , 1000) , 0101)" ∷
"('0' , '1') , (0000 , 0001) , 0000)" ∷ []
-}

{- pretty printed by hand in the figure.

((0 , 0) , (0000 , 0000) , 0000) ∷
((0 , 1) , (0000 , 0001) , 0000) ∷
((1 , 0) , (0001 , 0001) , 0001) ∷
((0 , 0) , (0001 , 0001) , 0001) ∷
((1 , 0) , (0001 , 0010) , 0001) ∷
((0 , 0) , (0001 , 0010) , 0001) ∷
((1 , 0) , (0010 , 0011) , 0010) ∷
((0 , 0) , (0010 , 0011) , 0010) ∷
((1 , 0) , (0011 , 0101) , 0011) ∷
((0 , 0) , (0011 , 0101) , 0011) ∷
((1 , 0) , (0101 , 1000) , 0101) ∷
((0 , 1) , (0000 , 0001) , 0000) ∷ []
-}


