module code.examples.STI where

-- revisit the imports
open import Agda.Builtin.String renaming (primStringAppend to _+++_)
open import Agda.Primitive
open import Data.Bool renaming (Bool to Bit; false to 0#; true to 1#)
open import Data.Fin using (Fin; zero; suc)
open import Data.List hiding (lookup ; [_] ; and)
open import Data.Maybe hiding (map)
open import Data.Nat using (ℕ; _∸_; _<_ ; _≤_ ; _+_) renaming (zero to Z; suc to S)
open import Data.Product hiding (map ; swap ; curry ; uncurry)
open import Data.Vec hiding (foldl ; foldr ; _++_ ; map ; _[_]=_ ; length) renaming (lookup to vecref)

open import code.common.Idioms hiding (_∘_)
open import code.common.Contexts
open import code.common.MachineWords
open import code.common.Stream renaming (hd to shd; tl to stl ; map to mapStr ; zip to zipStr)

open import code.firrtl.Types
open import code.firrtl.Syntax
open import code.firrtl.TransCircuit

open import code.devicecalculus.MealyMachine
open import code.devicecalculus.Semantics


-- Some abbreviations

W1 : Ty
W1 = ↑ UInt 1

---------------------
---------------------

Δ : ModTy
Δ = Mod (W1 ⊗ Unit) , (W1 ⊗ Unit) , (W1 ⊗ Unit)  ⋊ []

ost : [] ⊢ₓ UInt 1 // Δ -- out <- reg
ost = reg Zᵍ 

sst : [] ⊢ₓ UInt 1 // Δ
sst = mux (inp Zᵍ)
          (mux (reg Zᵍ) (uint (0# ∷ [])) (uint (1# ∷ [])))
          (reg Zᵍ)

sti : [] ⊢ᶜ Δ
sti = mod (mod ◇
             (node none (⊗case ost unit)) -- outs
             (node none (⊗case sst unit)) -- regs
             (node none unit))

{-
circuit sti :
  module sti :
    input i    : UInt<1>
    input clk   : Clock
    reg s       : UInt<1>, clk
    output o   : UInt<4>

    n <= if reset then 0 
         elsif go then m else n
    m <= if reset then 1
         elsif go then n + m else m
    out <= n
-}

sti' : Mealy (Vec Bit 1 × ⊤) (Vec Bit 1 × ⊤) (Vec Bit 1 × ⊤)
sti' = ⟦ trC Δ [] sti ⟧ []

insid : Stream (Vec Bit 1 × ⊤)
insid = cycle (0# ∷ [] , tt) ((1# ∷ [] , tt) ∷ (1# ∷ [] , tt) ∷ (1# ∷ [] , tt) ∷ [])

-- these are the interactive inputs.
inputs = map (λ (b , _) → ppBV b) (takeStr 20 insid)

triple : String → String → String → String
triple i s o = "(" +++ (i +++ (" , " +++ (s +++ (" , " +++ (o +++ ")")))))

-- hideous
pp : ((Vec Bit 1 × ⊤ {lzero}) × (Vec Bit 1 × ⊤ {lzero}) × (Vec Bit 1 × ⊤ {lzero})) → String
pp ((i , _) , (s , _) , (o , _)) = triple ppi pps ppo
  where
     ppi = ppBV i
     pps = ppBV s
     ppo = ppBV o

runsti = map pp (takeStr 20 (transcript sti' insid))

{-
"(0 , 0 , 0)" ∷
"(0 , 0 , 0)" ∷
"(1 , 1 , 0)" ∷
"(1 , 0 , 1)" ∷
"(1 , 1 , 0)" ∷
"(0 , 1 , 1)" ∷
"(1 , 0 , 1)" ∷
"(1 , 1 , 0)" ∷
"(1 , 0 , 1)" ∷
"(0 , 0 , 0)" ∷
"(1 , 1 , 0)" ∷
"(1 , 0 , 1)" ∷
"(1 , 1 , 0)" ∷
"(0 , 1 , 1)" ∷
"(1 , 0 , 1)" ∷
"(1 , 1 , 0)" ∷
"(1 , 0 , 1)" ∷ 
"(0 , 0 , 0)" ∷ 
"(1 , 1 , 0)" ∷ 
"(1 , 0 , 1)" ∷ []
-}
