module code.firrtl.TransType where

open import Relation.Binary.PropositionalEquality hiding ([_])

open import Data.Bool renaming (Bool to Bit; false to 0#; true to 1#)
open import Data.Fin using (Fin; zero; suc)
open import Data.List hiding (lookup ; [_])
open import Data.Maybe hiding (map)
open import Data.Nat using (ℕ; _∸_; _<_ ; _≤_ ; _+_) renaming (zero to Z; suc to S)
open import Data.Product hiding (map ; swap ; curry ; uncurry)
open import Data.Vec hiding (head ; tail ; foldl ; foldr ; _++_ ; map ; _[_]=_ ; length ; count) renaming (lookup to vecref)

open import code.common.Idioms
open import code.common.Contexts

open import code.firrtl.Types
open import code.firrtl.Syntax

open import code.devicecalculus.Types
open import code.devicecalculus.BitPatterns
open import code.devicecalculus.Syntax

Π : Type → Cxt Type → Type
Π h []       = h
Π h (t ∷ []) = h ⊗ t
Π h (t ∷ ts) = h ⊗ (Π t ts)

trTy : Ty → Type
trTy (↑ UInt n) = slv n
trTy (↑ SInt n) = slv n
trTy (↑ Clock)  = bit   -- don't care
trTy (t1 ⊗ t2)  = (trTy t1) ⊗ trTy t2
trTy Unit       = unit

lemma0 : (τ t : Ty) → (ts : List Ty) → trTy (Πty τ (t ∷ ts)) ≡ (trTy τ ⊗ trTy (Πty t ts))
lemma0 τ t []       = refl
lemma0 τ t (x ∷ ts) = refl

lemma1 : (τ t : Type) → (ts : List Type) → Π τ (t ∷ ts) ≡ (τ ⊗ (Π t ts))
lemma1 τ t []       = refl
lemma1 τ t (x ∷ ts) = refl

lemma2 : (τ : Ty) → (ts : List Ty) → Π (trTy τ) (map trTy ts) ≡ trTy (Πty τ ts)
lemma2 τ []       = refl
lemma2 τ (t ∷ ts) rewrite lemma0 τ t ts
                        | lemma1 (trTy τ) (trTy t) (map trTy ts)
                        | lemma2 t ts = refl

lemma3 : (τ : Ty) → (ts : List Ty) → trTy (Πty τ ts) ≡ Π (trTy τ) (map trTy ts)
lemma3 τ []       = refl
lemma3 τ (t ∷ ts) rewrite lemma0 τ t ts
                          | lemma1 (trTy τ) (trTy t) (map trTy ts)
                          | lemma3 t ts = refl

lemma4 : (τ t : Ty) → (ts : List Ty) → trTy (Πty τ (t ∷ ts)) ≡ (trTy τ ⊗ (Π (trTy t) (map trTy ts)))
lemma4 τ t []       = refl
lemma4 τ t (x ∷ ts) rewrite lemma3 t (x ∷ ts) = refl

--
-- Computing initial values for Dev. Given a (t : Ty),
-- need a (Γ ⊩ᵈ trTy t) which is all zeros
--
zeroVec : (n : ℕ) → Vec Bit n
zeroVec Z     = []
zeroVec (S n) = 0# ∷ zeroVec n

mk0 : ∀ {Γ} → (n : ℕ) → Γ ⊩ᵈ (slv n)
mk0 n = ⟨ zeroVec n ⟩

value0 : ∀ {Γ} → (t : Ty) → (Γ ⊩ᵈ (trTy t))
value0 (↑ UInt n) = mk0 n
value0 (↑ SInt n) = mk0 n
value0 (↑ Clock)  = 0!
value0 (t1 ⊗ t2)  = value0 t1 ⊠ value0 t2
value0 Unit       = 𝟙

mutual

  tystates : Cxt ModTy → Cxt Ty  
  tystates []      = []
  tystates (m ∷ μ) = tystate m ∷ (tystates μ)

  tystate : ModTy → Ty
  tystate (Mod _ , s , _ ⋊ []) = s
  tystate (Mod _ , s , _ ⋊ ms) = Πty s (tystates ms)

--
-- Given, Δ : Mod i , s , o ⋊ μ, these returns tystate Δ (state Δ)
--

stateTy : Ty → Cxt ModTy → Ty 
stateTy s [] = s
stateTy s μ  = Πty s (tystates μ)

stateType : Ty → Cxt ModTy → Type
stateType s μ = trTy (stateTy s μ)

--
-- Returns a term with the initial value for (state Δ)
--
state0 : ∀ {Γ} → (Δ : ModTy) → Γ ⊩ᵈ trTy (tystate Δ)
state0 Δ = value0 (tystate Δ)
